/*
 * Copyright (c) 2009-2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <portLib.h>
#include <h2timeLib.h>

#include <rmp440/rmp440.h>

#include "rmp440Const.h"
#include "server/rmp440Header.h"

#ifdef WITH_FELIB
#include <fe/ftdi-emergency.h>
#endif

#include "odoProba.h"
#include "joystick.h"

#ifdef GYRO
#include <gyroLib/gyro.h>
#endif

extern struct rmp440_feedback *data;

/*------------------------------------------------------------------------
 * Log
 *
 * Description:
 *
 * Reports:      OK
 *              S_rflex_CANNOT_OPEN_FILE
 *              S_rflex_CANNOT_WRITE_FILE
 */

static FILE *rmp440LogFile;

/* rmp440LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogStart(GEN_STRING128 *fileName, int *report)
{
	if (!(rmp440LogFile = fopen(fileName->name, "a"))) {
		printf ("rmp440LogStart: cannot open odometer file %s\n",
		    fileName->name);
		*report = S_rmp440_CANNOT_OPEN_FILE;
		return ETHER;
	}

	if (!fprintf(rmp440LogFile, "date\tdate1\tdate2\tv\tw\tmain/yaw\tmain/pitch\tmain/roll\tmain/x\tmain/y\tmain/z\tvlocal/yaw\tvlocal/pitch\tvlocal/roll\tvlocal/x\tvlocal/y\tvlocal/z\n")) {
		printf ("rmp440LogStart: cannot log in %s\n", fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInter(fileName, report);
		return ETHER;
	}

	printf ("rmp440LogStart: log file %s found\n", fileName->name);
	return EXEC;
}

/* rmp440LogExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogExec(GEN_STRING128 *fileName, int *report)
{
	H2TIMESPEC tv;

	/* get date */
	h2GetTimeSpec(&tv);

	if (!fprintf(rmp440LogFile, "%lf\t%d\t%d\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\n",
		tv.tv_sec + tv.tv_nsec*1.0E-9, SDI_F->mePos.date1, SDI_F->mePos.date2,
		SDI_F->robot.v, SDI_F->robot.w,
		SDI_F->mePos.main.euler.yaw, SDI_F->mePos.main.euler.pitch, SDI_F->mePos.main.euler.roll,
		SDI_F->mePos.main.euler.x, SDI_F->mePos.main.euler.y, SDI_F->mePos.main.euler.z,
		SDI_F->mePos.VLocal.euler.yaw, SDI_F->mePos.VLocal.euler.pitch, SDI_F->mePos.VLocal.euler.roll,
		SDI_F->mePos.VLocal.euler.x, SDI_F->mePos.VLocal.euler.y, SDI_F->mePos.VLocal.euler.z
		)) {
		printf ("rmp440LogExec: cannot log in %s\n", fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInter(fileName, report);
		return ETHER;
	}

	return EXEC;
}


/* rmp440LogInter  -  codel INTER of Log
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInter(GEN_STRING128 *fileName, int *report)
{
	if (rmp440LogFile) {
		fclose(rmp440LogFile);
		rmp440LogFile = NULL;
	}
	return ETHER;
}

/*------------------------------------------------------------------------
 * LogInternalData
 *
 * Description: Log all the raw data provided by robot
 *
 * Reports:     OK
 *              S_rmp440_CANNOT_OPEN_FILE
 *              S_rmp440_CANNOT_WRITE_FILE
 */

static FILE *rmp440LogInternalDataFile;

/* rmp440LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataStart(GEN_STRING128 *fileName, int *report)
{
	if (!(rmp440LogInternalDataFile = fopen(fileName->name, "a"))) {
		printf ("rmp440LogInternalDataStart: cannot open "
		    "internal data file %s\n", fileName->name);
		*report = S_rmp440_CANNOT_OPEN_FILE;
		return ETHER;
  }

	if (!fprintf(rmp440LogInternalDataFile,
		"date\tdate1\tdate2"
		"\tgyroAngle\tgyroRate"
		"\tsent_velocity_reference\tsent_velocity_command"
		"\tsent_turn_reference\tsent_turn_command"
		"\tstraight\tstraight_angle"
		/* begin of rmp440 feedback */
		"\tfault_status0\tfault_status1\tfaultstatus2\tfault_status3"
		"\tmcu_fault0\tmcu_fault1\tmcu_fault2\tmcu_fault3"
		"\tframe_count\toperational_state\tdynamic_response"
		"\tmin_propulstion_bat_soc\taux_batt_soc"
		"\tinertial_x_acc\tinertial_y_acc\tinertial_x_rate"
		"\tinertial_y_rate\tinertial_z_rate"
		"\tpse_pitch\tpse_pitch_rate\tpse_roll\tpse_roll_rate"
		"\tpse_yaw_rate\tpse_data_is_valid"
		"\tyaw_rate_limit\tvel_limit\tlinear_accel\tlinear_vel"
		"\tdifferential_wheel_vel\tright_front_vel\tleft_front_vel"
		"\tright_rear_vel\tleft_rear_vel"
		"\tright_front_pos\tleft_front_pos"
		"\tright_rear_pos\tleft_rear_pos\tlinear_pos"
		"\tright_front_current\tleft_front_current"
		"\tright_rear_current\tleft_rear_current\tmax_motor_current"
		"\tright_front_current_limit\tleft_front_current_limit"
		"\tright_rear_current_limit\tleft_rear_current_limit"
		"\tmin_motor_current_limit"
		"\tfront_base_batt_1_soc\tfront_base_batt_2_soc"
		"\trear_base_batt_1_soc\trear_base_batt_2_soc"
		"\tfront_base_batt_1_temp\tfront_base_batt_2_temp"
		"\trear_base_batt_1_temp\trear_base_batt_2_temp"
		"\tvel_target\tyaw_rate_target\tangle_target"
		"\taux_batt_voltage\taux_batt_current\taux_batt_temp"
		"\tabb_system_status\tabb_batt_status\taux_batt_faults"
		"\tccu_7p2_battery_voltage\tsp_sw_build_id\tuip_sw_build_id"
		"\tmcu_inst_power0\tmcu_inst_power1\tmcu_inst_power2"
		"\tmcu_inst_power3"
		"\tmcu_total_energy0\tmcu_total_energy1"
		"\tmcu_total_energy2\tmcu_total_energy3\n")) {
		printf ("rmp440LogStart: cannot log in %s\n", fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInternalDataInter(fileName, report);
		return ETHER;
	}
	printf ("rmp440LogInternalDataStart: log file %s found\n", fileName->name);
	return EXEC;
}

/* rmp440LogInternalDataExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataExec(GEN_STRING128 *fileName, int *report)
{
	H2TIMESPEC tv;

	/* get date */
	h2GetTimeSpec(&tv);
#if 1
	if (!fprintf(rmp440LogInternalDataFile, "%lf\t%d\t%d"
		"\t%.6g\t%.6g"
		"\t%.6g\t%.6g"
		"\t%.6g\t%.6g"
		"\t%d\t%.6g"
		/* start of feedback */
		"\t0x%08x\t0x%08x\t0x%08x\t%08x"
		"\t0x%08x\t0x%08x\t0x%08x\t%08x"
		"\t%.6g\t%d\t%d\t%.6g\t%.6g"
		"\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\t%d\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\t0x%08x\t0x%08x\t0x%08x\t%.6g\t0x%08x\t0x%08x"
		"\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g"
		"\n",
		/* timestamps */
		tv.tv_sec + tv.tv_nsec*1.0E-9,
		SDI_F->mePos.date1, SDI_F->mePos.date2,
		/* the KVH gyro */
		SDI_F->gyro.gyroTheta,
		SDI_F->gyro.gyroOmega,
		SDI_F->vReference, SDI_F->vCommand, // need to check if latest value is logged
		SDI_F->wReference, SDI_F->wCommand,
		SDI_F->gyroAsserv.straight, SDI_F->gyroAsserv.straight_angle,
		/* rmp440 feedback */
		data->fault_status[0], data->fault_status[1],
		data->fault_status[2], data->fault_status[3],
		data->mcu_fault_status[0], data->mcu_fault_status[1],
		data->mcu_fault_status[2], data->mcu_fault_status[3],
		data->frame_count,
		data->operational_state,
		data->dynamic_response,
		data->min_propulsion_batt_soc,
		data->aux_batt_soc,
		data->inertial_x_acc,
		data->inertial_y_acc,
		data->inertial_x_rate,
		data->inertial_y_rate,
		data->inertial_z_rate,
		data->pse_pitch,
		data->pse_pitch_rate,
		data->pse_roll,
		data->pse_roll_rate,
		data->pse_yaw_rate,
		data->pse_data_is_valid,
		data->yaw_rate_limit,
		data->vel_limit,
		data->linear_accel,
		data->linear_vel,
		data->differential_wheel_vel,
		data->right_front_vel,
		data->left_front_vel,
		data->right_rear_vel,
		data->left_rear_vel,
		data->right_front_pos,
		data->left_front_pos,
		data->right_rear_pos,
		data->left_rear_pos,
		data->linear_pos,
		data->right_front_current,
		data->left_front_current,
		data->right_rear_current,
		data->left_rear_current,
		data->max_motor_current,
		data->right_front_current_limit,
		data->left_front_current_limit,
		data->right_rear_current_limit,
		data->left_rear_current_limit,
		data->min_motor_current_limit,
		data->front_base_batt_1_soc,
		data->front_base_batt_2_soc,
		data->rear_base_batt_1_soc,
		data->rear_base_batt_2_soc,
		data->front_base_batt_1_temp,
		data->front_base_batt_2_temp,
		data->rear_base_batt_1_temp,
		data->rear_base_batt_2_temp,
		data->vel_target,
		data->yaw_rate_target,
		data->angle_target,	/* only used on omni platforms */
		data->aux_batt_voltage,
		data->aux_batt_current,
		data->aux_batt_temp,
		data->abb_system_status,
		data->abb_batt_status,
		data->aux_batt_faults,
		data->ccu_7p2_battery_voltage,
		data->sp_sw_build_id,
		data->uip_sw_build_id,
		data->mcu_inst_power[0], data->mcu_inst_power[1],
		data->mcu_inst_power[2], data->mcu_inst_power[3],
		data->mcu_total_energy[0], data->mcu_total_energy[1],
		data->mcu_total_energy[2], data->mcu_total_energy[3]))
	{
		printf ("rmp440LogInternalDataExec: cannot log in %s\n",
		    fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInter(fileName, report);
		return ETHER;
	}
#endif
	return EXEC;
}

/* rmp440LogInternalDataInter  -  codel INTER of LogInternalData
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataInter(GEN_STRING128 *fileName, int *report)
{
	if (rmp440LogInternalDataFile) {
		fclose(rmp440LogInternalDataFile);
		rmp440LogInternalDataFile = NULL;
	}
	return ETHER;
}
